
import os
import json
from pybythec import utils
from pybythec.utils import f

log = utils.Logger('pybythec')

class BuildStatus:
  '''
    member variables:
    name: name of target
    path: path to write status.json file to
    status: failed, built, up to date, or locked
    description: what happened
  '''
                                     
  def __init__(self, name, path = ''):
    self.name = name
    self.path = path
    self.status = 'failed'
    self.description = ''
  

  def readFromFile(self, libSrcDir, buildDir, buildType, compiler, binaryFormat):
    '''
      buildPath (in): where to read the status.json file from
    '''

    buildPath = f('{0}/{1}/{2}/{3}/{4}', libSrcDir, buildDir, buildType, compiler, binaryFormat)
    if not os.path.exists(buildPath):
      # try the other hidden / non-hidden version
      if buildDir[0] == '.':
        buildPath = f('{0}/{1}/{2}/{3}/{4}', libSrcDir, buildDir.lstrip('.'), buildType, compiler, binaryFormat)
      else:
        buildPath = f('{0}/.{1}/{2}/{3}/{4}', libSrcDir, buildDir, buildType, compiler, binaryFormat)

    contents = utils.loadJsonFile(buildPath + '/status.json')
    if not contents:
      self.description = 'couldn\'t find contents in ' + buildPath
      log.error('couldn\'t find contents in ' + buildPath)
      return
    if 'status' in contents:
      self.status = contents['status']
    else:
      self.description = 'couldn\'t find the build status in ' + buildPath
      log.error(self.description)
    if 'description' in contents:
      self.description = contents['description']
    else:
      self.description = buildPath + ' doesn\'t contain a description'
      log.warning(self.description)


  def writeInfo(self, status, msg, *args):
    msg = f(msg, *args)
    log.info(msg)
    self.status = status
    self.description = msg
    self._writeToFile()


  def writeError(self, msg, *args):
    msg = f(msg, *args)
    log.error(msg)
    self.description = msg
    self._writeToFile()


  def _writeToFile(self):
    if not os.path.exists(self.path):
      return
    with open(self.path + '/status.json', 'w') as f:
      json.dump({'status': self.status, 'description': self.description}, f, indent = 4)



# -*- coding: utf-8 -*-
from pybythec import utils
from pybythec.utils import f
from pybythec.utils import PybythecError
from pybythec.BuildStatus import BuildStatus
from pybythec.BuildElements import BuildElements

import os
import sys
import time
from threading import Thread

log = utils.Logger('pybythec')

__author__ = 'glowtree'
__email__ = 'tom@glowtree.com'
__version__ = '0.9.43'


def getBuildElements(osType = None,
                     compiler = None,
                     buildType = None,
                     binaryFormat = None,
                     projConfigPath = None,
                     globalConfigPath = None,
                     projConfig = None,
                     globalConfig = None,
                     libDir = None):
  '''
    passthrough function that catches and reports exceptions
  '''
  try:
    return BuildElements(
        osType = osType,
        compiler = compiler,
        buildType = buildType,
        binaryFormat = binaryFormat,
        projConfig = projConfig,
        projConfigPath = projConfigPath,
        globalConfig = globalConfig,
        globalConfigPath = globalConfigPath,
        libDir = libDir)
  except PybythecError as e:
    log.error(e)
    return None
  except Exception as e:
    log.error('unknown exception: {0}', e)
    return None


def build(be = None, builds = None):
  '''
    be: BuildElements object
    builds: list of build overrides
  '''
  if not be:
    be = getBuildElements()
    if not be:
      return

  buildsRef = builds
  if not buildsRef:
    buildsRef = be.builds
  if type(buildsRef) is not list:
    buildsRef = [buildsRef]

  for build in buildsRef:
    try:
      be.configBuild(buildName = build)
    except PybythecError as e:
      log.error(e)
      continue
    except Exception as e:
      log.error('unknown exception: {0}', e)
      continue
    _build(be)


def _build(be):
  '''
    does the dirty work of compiling and linking based on the state setup in the BuildElements object be
  '''
  threading = True # NOTE: perhaps this could be an function argument

  buildStatus = BuildStatus(be.targetFilename, be.buildPath)

  # lock - early return
  if be.locked and os.path.exists(be.targetInstallPath):
    buildStatus.writeInfo('locked', '{0} is locked', be.targetName)
    return True

  startTime = time.time()

  log.info('building ' + be.infoStr)

  buildingLib = False
  if be.libDir:
    buildingLib = True

  if not os.path.exists(be.installPath):
    utils.createDirs(be.installPath)

  if not os.path.exists(be.buildPath):
    os.makedirs(be.buildPath)

  incPathList = []
  for incPath in be.incPaths:
    if os.path.exists(incPath):
      incPathList += ['-I', incPath]
    else:
      log.warning('incPath {0} doesn\'t exist', incPath)

  for extIncPath in be.extIncPaths:  # external include libs (for cases where 3rd party header includes are using "" instead of <> ie Unreal)
    if os.path.exists(incPath):
      incPathList += ['-I', extIncPath]
    else:
      log.warning('extIncPath {0} doesn\'t exist', extIncPath)

  definesList = []
  for define in be.defines:
    definesList += ['-D', define]

  #
  # qt moc file compilation
  #
  mocPaths = []
  for qtClass in be.qtClasses:
    found = False
    mocPath = f('{0}/moc_{1}.cpp', be.buildPath, qtClass)
    qtClassHeader = qtClass + '.h'

    for incPath in be.incPaths:  # find the header file, # TODO: should there be a separate list of headers ie be.mocIncPaths?
      includePath = incPath + '/' + qtClassHeader
      if not os.path.exists(includePath):
        continue

      if os.path.exists(mocPath) and float(os.stat(mocPath).st_mtime) < float(os.stat(includePath).st_mtime) or not os.path.exists(mocPath):
        buildStatus.description = 'qt moc: ' + utils.runCmd(['moc'] + definesList + [includePath, '-o', mocPath])

      if not os.path.exists(mocPath):
        buildStatus.writeError(buildStatus.description)
        return False

      mocPaths.append(mocPath)
      found = True

    if not found:
      buildStatus.writeError('can\'t find {0} for qt moc compilation', qtClassHeader)
      return False

  for mocPath in mocPaths:
    be.sources.append(mocPath)

  buildStatusDeps = []  # the build status for each dependency: objs and libs
  threads = []
  i = 0

  #
  # compile
  #
  objPaths = []
  cmd = [be.compilerCmd, be.objFlag] + incPathList + definesList + be.flags

  if threading:
    for source in be.sources:
      buildStatusDep = BuildStatus(source)
      buildStatusDeps.append(buildStatusDep)
      thread = Thread(None, target = _compileSrc, args = (be, cmd, source, objPaths, buildStatusDep))
      thread.start()
      threads.append(thread)
      i += 1
  else:
    for source in be.sources:
      buildStatusDep = BuildStatus(source)
      buildStatusDeps.append(buildStatusDep)
      _compileSrc(be, cmd, source, objPaths, buildStatusDep)
      i += 1

  #
  # build library dependencies
  #
  libCmds = []
  libsBuilding = []
  if be.binaryType == 'exe' or be.binaryType == 'plugin':
    for lib in be.libs:
      libName = lib
      if be.compiler.startswith('msvc'):
        libCmds += [libName + be.staticExt]  # you need to link against the .lib stub file even if it's ultimately a .dll that gets linked
      else:
        libCmds += [be.libFlag, libName]

      # check if the lib has a directory for building
      if threading:
        for libSrcDir in be.libSrcPaths:
          libSrcDir = os.path.join(libSrcDir, lib)
          if os.path.exists(libSrcDir):
            libsBuilding.append(lib)
            buildStatusDep = BuildStatus(lib)
            buildStatusDeps.append(buildStatusDep)
            thread = Thread(None, target = _buildLib, args = (be, libSrcDir, buildStatusDep))
            thread.start()
            threads.append(thread)
            i += 1
            break
      else:
        for libSrcPath in be.libSrcPaths:
          if not os.path.exists('libSrcPath'):
            log.warning('libSrcPath {0} doesn\'t exist', libSrcPath)
            continue
          libSrcPath = os.path.join(libSrcPath, lib)
          if os.path.exists(libSrcPath):
            libsBuilding.append(lib)
            buildStatusDep = BuildStatus(lib)
            buildStatusDeps.append(buildStatusDep)
            _buildLib(be, libSrcDir, buildStatusDep)
            i += 1
            break

  # wait for all the threads before checking the results
  for thread in threads:
    thread.join()

  allUpToDate = True
  for buildStatusDep in buildStatusDeps:
    if buildStatusDep.status == 'failed':
      # NOTE: changed from buildStatusDep.description.encode('ascii', 'ignore') which fixed issue on macOs
      buildStatus.writeError('{0} failed because {1} failed because...\n\n{2}\n...determined in seconds\n\n', be.infoStr, buildStatusDep.name,
                             buildStatusDep.description, str(int(time.time() - startTime)))
      return False
    elif buildStatusDep.status == 'built':
      allUpToDate = False

  # revise the library paths
  for i in range(len(be.libPaths)):
    revisedLibPath = be.libPaths[i] + be.binaryRelPath
    if os.path.exists(revisedLibPath):
      be.libPaths[i] = revisedLibPath

  # check for multiple instances of a lib: link erros due to linking to the wrong version of a lib can be a nightmare to debug
  # if you don't suspect it's the wrong version
  libsFound = {} # lib name, array of paths where it was found
  for p in be.libPaths:
    for lib in be.libs:
      if be.compiler.startswith('msvc'):
        staticPath = f('{0}/{1}{2}', p, lib, be.staticExt)
        dynamicPath = f('{0}/{1}{2}', p, lib, be.dynamicExt)
      else:
        staticPath = f('{0}/lib{1}{2}', p, lib, be.staticExt)
        dynamicPath = f('{0}/lib{1}{2}', p, lib, be.dynamicExt)
      if os.path.exists(staticPath) or os.path.exists(dynamicPath):
        if lib in libsFound:
          libsFound[lib].append(p)
        else:
          libsFound[lib] = [p]
  for l in libsFound:
    libPaths = libsFound[l]
    if len(libPaths) > 1:
      log.w('lib {0} found in more than one place: {1}\n', l, libPaths)

  #
  # linking
  #
  linkCmd = []

  if allUpToDate and os.path.exists(be.targetInstallPath):
    buildStatus.writeInfo('up to date', '{0} is up to date, determined in {1} seconds\n', be.infoStr, str(int(time.time() - startTime)))
    if not buildingLib:
      _runPostScript(be)
    return True

  # microsoft's compiler / linker can only handle so many characters on the command line
  msvcLinkCmdFilePath = be.buildPath + '/linkCmd'
  if be.compiler.startswith('msvc'):
    msvcLinkCmd = f('{0}"{1}" "{2}" {3}', be.targetFlag, be.targetInstallPath, '" "'.join(objPaths), ' '.join(libCmds))
    msvcLinkCmdFp = open(msvcLinkCmdFilePath, 'w')
    msvcLinkCmdFp.write(msvcLinkCmd)
    msvcLinkCmdFp.close()
    linkCmd += [be.linker, '@' + msvcLinkCmdFilePath]
    if be.showLinkerCmds:
      log.info('\nmsvcLinkCmd: {0}\n', msvcLinkCmd)
  else:
    linkCmd += [be.linker, be.targetFlag, be.targetInstallPath] + objPaths + libCmds

  if be.binaryType != 'static':  # TODO: is this the case for msvc?
    linkCmd += be.linkFlags

  if be.binaryType == 'exe' or be.binaryType == 'plugin' or (be.compilerRoot == 'msvc' and be.binaryType == 'dynamic'):

    for libPath in be.libPaths:
      if not os.path.exists(libPath):
        log.warning('libPath {0} doesn\'t exist', libPath)
        continue
      if be.compiler.startswith('msvc'):
        linkCmd += [be.libPathFlag + os.path.normpath(libPath)]
      else:
        linkCmd += [be.libPathFlag, os.path.normpath(libPath)]

  # get the timestamp of the existing target if it exists
  linked = False
  targetExisted = False
  oldTargetTimeStamp = None
  if os.path.exists(be.targetInstallPath):
    oldTargetTimeStamp = float(os.stat(be.targetInstallPath).st_mtime)
    targetExisted = True

  if be.showLinkerCmds:
    log.info('\n{0}\n', ' '.join(linkCmd))

  buildStatus.description = utils.runCmd(linkCmd)

  if os.path.exists(be.targetInstallPath):
    if targetExisted:
      if float(os.stat(be.targetInstallPath).st_mtime) > oldTargetTimeStamp:
        linked = True
    else:
      linked = True

  if linked:
    log.info('linked ' + be.infoStr)
  else:
    buildStatus.writeError('linking failed because {0}', buildStatus.description)
    return False

  # copy dynamic library dependencies (built by this build) to the install path
  if be.binaryType == 'exe' or be.binaryType == 'plugin':
    for lib in be.libs:
      for libPath in be.libPaths:
        dynamicPath = libPath + '/'
        if be.compilerRoot == 'gcc' or be.compilerRoot == 'clang':
          dynamicPath += 'lib'
        dynamicPath += lib + be.dynamicExt
        if os.path.exists(dynamicPath):
          utils.copyfile(dynamicPath, be.installPath)

  buildStatus.writeInfo('built', '{0} built {1}\ncompleted in {2} seconds\n', be.infoStr, be.targetInstallPath, str(int(time.time() - startTime)))

  sys.stdout.flush()

  # run a post-build script if it exists
  if not buildingLib:
    _runPostScript(be)

  return True


#
# private functions
#
def _compileSrc(be, compileCmd, source, objPaths, buildStatus):
  '''
    be (in): BuildElements object
    compileCmd (in): the compile command so far
    source (in): the c or cpp source file to compile (every source file gets it's own object file)
    objPaths (out): list of all object paths that will be passed to the linker
    buildStatus (out): build status for this particular compile, defaults to failed
  '''

  if not os.path.exists(source):
    buildStatus.writeError('{0} is missing, exiting build', source)
    return

  objFile = os.path.basename(source)
  objFile = objFile.replace(os.path.splitext(source)[1], be.objExt)
  objPath = os.path.join(be.buildPath, objFile)
  objPaths.append(objPath)

  # check if it's up to date
  objExisted = os.path.exists(objPath)
  if objExisted:
    objTimestamp = float(os.stat(objPath).st_mtime)
    if not utils.sourceNeedsBuilding(be.incPaths, source, objTimestamp):
      buildStatus.status = 'up to date'
      return

  # Microsoft Visual C has to have the objPathFlag cuddled up directly next to the objPath - no space in between them (grrr)
  if be.compiler.startswith('msvc'):
    cmd = compileCmd + [source, be.objPathFlag + objPath]
  else:
    cmd = compileCmd + [source, be.objPathFlag, objPath]

  if be.showCompilerCmds:
    log.info('\n' + ' '.join(cmd) + '\n')

  buildStatus.description = utils.runCmd(cmd)

  if os.path.exists(objPath):
    if objExisted:
      if float(os.stat(objPath).st_mtime) > objTimestamp:
        buildStatus.status = 'built'
    else:
      buildStatus.status = 'built'

  if buildStatus.status == 'built':
    buildStatus.description = 'compiled ' + os.path.basename(source)
  else:
    log.error('{0} failed to build', objPath)


def _buildLib(be, libSrcDir, buildStatus):
  '''
  '''
  jsonPath = os.path.join(libSrcDir, 'pybythec.json')
  if not os.path.exists(jsonPath):
    jsonPath = os.path.join(libSrcDir, '.pybythec.json')
  if not os.path.exists(jsonPath):
    buildStatus.writeError('{0} does not have a pybythec.json file', libSrcDir)
    return

  libBe = getBuildElements(
      osType = be.osType,
      compiler = be.compiler,
      buildType = be.buildType,
      binaryFormat = be.binaryFormat,
      projConfig = be.projConfig,
      globalConfig = be.globalConfig,
      libDir = libSrcDir)
  if not libBe:
    return
  build(libBe)

  # read the build status
  buildStatus.readFromFile(libSrcDir, be.buildDir, be.buildType, be.compiler, be.binaryFormat)


def clean(be = None, builds = None):
  '''
  '''
  if not be:
    be = getBuildElements()
    if not be:
      return

  buildsRef = builds
  if not buildsRef:
    buildsRef = be.builds
  if type(buildsRef) is not list:
    buildsRef = [buildsRef]

  for build in buildsRef:
    try:
      be.configBuild(buildName = build)
    except PybythecError as e:
      log.error(e)
      return
    except Exception as e:
      log.error('unknown exception: {0}', e)
      return
    _clean(be)


def _clean(be = None):
  '''
    cleans the current project
    be (in): BuildElements object
  '''

  # remove any dynamic libs that are sitting next to the exe
  if os.path.exists(be.installPath) and (be.binaryType == 'exe' or be.binaryType == 'plugin'):
    for fl in os.listdir(be.installPath):
      libName, ext = os.path.splitext(fl)
      if ext == be.dynamicExt:
        if be.compilerRoot == 'gcc' or be.compilerRoot == 'clang':
          libName = libName.lstrip('lib')
        for lib in be.libs:
          if lib == libName:
            p = be.installPath + '/' + fl
            try:
              os.remove(p)
            except Exception:
              log.warning('failed to remove ' + p)
      elif ext == '.exp' or ext == '.ilk' or ext == '.lib' or ext == '.pdb':  # msvc files
        p = be.installPath + '/' + fl
        try:
          os.remove(p)
        except Exception:
          log.warning('failed to remove ' + p)

  if not os.path.exists(be.buildPath):  # canary in the coal mine
    log.info(be.infoStr + ' already clean')
    return True

  dirCleared = True
  for fl in os.listdir(be.buildPath):
    p = be.buildPath + '/' + fl
    try:
      os.remove(p)
    except Exception:
      dirCleared = False
      log.warning('failed to remove {0}', p)
  if dirCleared:
    os.removedirs(be.buildPath)

  if os.path.exists(be.targetInstallPath):
    os.remove(be.targetInstallPath)
  target, ext = os.path.splitext(be.targetInstallPath)
  if ext == '.dll':
    try:
      os.remove(target + '.exp')
      os.remove(target + '.lib')
    except Exception:
      pass
  try:
    os.removedirs(be.installPath)
  except Exception:
    pass

  log.info(be.infoStr + ' all clean')
  return True


def cleanAll(be = None, builds = None):
  '''
    cleans both the current project and also the dependencies
  '''
  if not be:
    be = getBuildElements()
    if not be:
      return

  buildsRef = builds
  if not buildsRef:
    buildsRef = be.builds
  if type(buildsRef) is not list:
    buildsRef = [buildsRef]

  for build in buildsRef:
    try:
      be.configBuild(buildName = build)
    except PybythecError as e:
      log.error(e)
      continue
    except Exception as e:
      log.error('unknown exception: {0}', e)
      continue
    _clean(be)
    # clean library dependencies
    for lib in be.libs:
      for libSrcPath in be.libSrcPaths:
        libPath = os.path.join(libSrcPath, lib)
        if os.path.exists(libPath):
          libBe = getBuildElements(
              osType = be.osType,
              compiler = be.compiler,
              buildType = be.buildType,
              binaryFormat = be.binaryFormat,
              projConfig = be.projConfig,
              globalConfig = be.globalConfig,
              libDir = libPath)
          if not libBe:
            return
          clean(libBe)  # builds = build)


def _runPostScript(be):
  '''
    looks for a post-build script and loads it as a module
  '''
  postScriptPath = './pybythecPost.py'
  if not os.path.exists(postScriptPath):
    postScriptPath = './.pybythecPost.py'
  if os.path.exists(postScriptPath):
    import imp
    m = imp.load_source('', postScriptPath)
    m.run(be)




import pybythec
import argparse


log = pybythec.utils.Logger('pybythec')

def main():

  parser = argparse.ArgumentParser(formatter_class = argparse.RawTextHelpFormatter)
  parser.add_argument('-v', '--version', action = 'store_true', help = 'the version')
  parser.add_argument('-c', '--compiler', help = 'any variation of gcc, clang, or msvc ie g++-4.4, msvc-110')  # metavar = 'compiler'
  parser.add_argument('-os', '--osType', help = 'operating system: currently linux, macOs, or windows')
  parser.add_argument('-b', '--buildType', help = 'debug release etc')
  parser.add_argument('-bf', '--binaryFormat', help = '32bit, 64bit etc')
  parser.add_argument('-pc', '--projectConfig', help = 'path to a pybythec project config file (json)')
  parser.add_argument('-gc', '--globalConfig', help = 'path to a pybythec global config file (json)')
  parser.add_argument('-bd', '--builds', help = 'list of builds: each one generates a seperate build (comma delineated, no spaces ie foo,bar)')
  parser.add_argument('-cl', '--clean', action = 'store_true', help = 'clean the build')
  parser.add_argument('-cla', '--cleanAll', action = 'store_true', help = 'clean the build and the builds of all library dependencies')
  args = parser.parse_args()

  if args.version:
    print('version: ' + pybythec.__version__)
    return

  be = pybythec.getBuildElements(
      osType = args.osType,
      compiler = args.compiler,
      buildType = args.buildType,
      binaryFormat = args.binaryFormat,
      projConfigPath = args.projectConfig,
      globalConfigPath = args.globalConfig)
  if not be:
    return

  builds = args.builds.split(',') if args.builds else None

  if args.cleanAll:
    return pybythec.cleanAll(be, builds)
  elif args.clean:
    return pybythec.clean(be, builds)
  else:
    return pybythec.build(be, builds)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

'''
test_pybythec
----------------------------------

tests for pybythec module
'''

import os
import platform
import unittest
import subprocess
import pybythec
from pybythec.utils import f

log = pybythec.utils.Logger('pybythecTest')

class TestPybythec(unittest.TestCase):
  
  def setUp(self):
    '''
      typical setup for building with pybythec
    '''
    # setup the environment variables...
    # normally you would probably set these in your .bashrc (linux / macOs), profile.ps1 (windows) file etc
    os.environ['PYBYTHEC_EXAMPLE_SHARED'] = '../../shared'
    
    self.builds = ['buildVar1', 'buildVar3'] # corresponds to build vartions declared in example/projects/Main/pybythec.json

    # make sure we have a .pybythecGlobals.json in the home directory
    globalsPath = os.path.expanduser('~') + '/.pybythecGlobals.json'
    if not os.path.exists(globalsPath):
      log.info('creating {0}', globalsPath)
      with open(globalsPath, 'w') as wf:
        with open('./globals.json') as rf:
          wf.write(rf.read())
    self.assertTrue(os.path.exists(os.path.expanduser('~') + '/.pybythecGlobals.json'))


  def test_000_something(self):
    '''
      build
    '''
    print('\n')

    # build Plugin
    os.chdir('./example/projects/Plugin')
    pybythec.build()
    
    # build Main (along with it's library dependencies)
    os.chdir('../Main')

    pybythec.build()
    
    for b in self.builds:
      exePath = f('./{0}/Main', b)
      if platform.system() == 'Windows':
        exePath += '.exe'
      
      self.assertTrue(os.path.exists(exePath))
      
      p = subprocess.Popen([exePath], stdout = subprocess.PIPE, stderr = subprocess.PIPE)
      stdout, stderr = p.communicate()
      stdout = stdout.decode('utf-8')
      log.info(stdout)
      
      if len(stderr):
        raise Exception(stderr)
      
      self.assertTrue(stdout.startswith('running an executable and a statically linked library and a dynamically linked library'))# and a plugin'))


  def tearDown(self):
    '''
      clean the builds
    '''
    pybythec.cleanAll()
    
    os.chdir('../Plugin')
    pybythec.cleanAll()


if __name__ == '__main__':
  import sys
  sys.exit(unittest.main())



from __future__ import print_function
import os
import sys
import json
import shutil
import subprocess


class PybythecError(Exception):
  def __init__(self, msg, *args):
    super(PybythecError, self).__init__(msg.format(*args))


def f(s, *args):
  return s.format(*args)


class Logger():

  wf = None  # static

  def __init__(self, name = None, debug = False):
    self.name = name
    if self.name:
      self.name += ': '
    else:
      self.name = ''
    self._debug = debug

  def setDebug(self, v):
    self._debug = v

  @classmethod
  def setFilepath(cls, filepath):
    if filepath:
      cls.wf = open(filepath, 'w')

  def _getStr(self, s, *args):
    if type(s) is not str:
      s = f('{0}', s)
    if len(args):
      return self.name + f(s, *args)
    else:
      return self.name + s

  def debug(self, s, *args):
    if self._debug:
      print('debug: ' + self._getStr(s, *args), file = Logger.wf)
      if Logger.wf:
        Logger.wf.flush()  # necessary for running as a systemd service for some reason

  def info(self, s, *args):
    print(self._getStr(s, *args), file = Logger.wf)
    if Logger.wf:
      Logger.wf.flush()

  def warning(self, s, *args):
    print('warning: ' + self._getStr(s, *args), file = Logger.wf)
    if Logger.wf:
      Logger.wf.flush()

  def error(self, s, *args):
    lwf = Logger.wf
    if not lwf:
      lwf = sys.stderr
    print('error: ' + self._getStr(s, *args), file = lwf)
    if Logger.wf:
      Logger.wf.flush()

  def raw(self, s, *args):  # no adding the name
    if type(s) is not str:
      s = f('{0}', s)
    if len(args):
      print(f(s, *args))
    else:
      print(s)

  # shorthands
  def d(self, s, *args):
    self.debug(s, *args)

  def i(self, s, *args):
    self.info(s, *args)

  def w(self, s, *args):
    self.warning(s, *args)

  def e(self, s, *args):
    self.error(s, *args)

  def r(self, s, *args):
    self.raw(s, *args)


log = Logger('pybythec')


def srcNewer(srcPath, dstPath):
  if int(os.stat(srcPath).st_mtime) > int(os.stat(dstPath).st_mtime):
    return True
  return False


def checkTimestamps(incPaths, src, timestamp):
  '''
    finds the newest timestamp of everything upstream of the src file, including the src file
  '''
  if not os.path.exists(src):
    log.warning('checkTimestamps: {0} doesn\'t exist', src)
    return

  srcTimeStamp = float(os.stat(src).st_mtime)
  if srcTimeStamp > timestamp[0]:
    timestamp[0] = srcTimeStamp

  fileCopy = str()
  srcFile = open(src, 'r')
  for line in srcFile:
    fileCopy += line
  srcFile.close()

  for line in fileCopy.split('\n'):
    if line.startswith('#include'):
      filename = line.lstrip('#include')
      filename = filename.strip()
      if (filename[0] == '"'):
        filename = filename.strip('"')
        for dir in incPaths:
          filepath = os.path.join(dir, filename)
          if os.path.exists(filepath):
            checkTimestamps(incPaths, filepath, timestamp)


def sourceNeedsBuilding(incPaths, src, objTimestamp):
  '''
    determines whether a source file needs to be built or not
  '''
  timestamp = [0]  # [] so it's passed as a reference
  checkTimestamps(incPaths, src, timestamp)

  if timestamp[0] > objTimestamp:
    return True

  return False


def getLibPath(libName, libPath, compiler, libExt):
  '''
      get the lib path with the os / compiler specific prefix and file extension
  '''
  libPath += '/'
  if compiler.startswith('gcc') or compiler.startswith('clang'):
    libPath += 'lib'
  libPath += libName + libExt
  return libPath


def makePathAbsolute(absPath, path):
  '''
    make a relative file path absolute
  '''
  if os.path.isabs(path):
    return path
  return os.path.normpath(os.path.join(absPath, './' + path))


def createDirs(path):
  '''
   recursively goes up the path heiarchy creating the necessary directories along the way
   similar to os.makedirs except doesn't throw an exception if a directory's already exists
   also os.makedirs throws the same exception whether the directory already exists or it couldn't create it, not ideal
  '''
  if path is None or not len(path):
    log.warning('createDirs: empty path')
    return

  # in case path ends with a '/'
  path = path.rstrip('/')

  if os.path.exists(path):
    return

  # if the path above the current one doesn't exist, create it
  abovePath = os.path.dirname(path)
  if not os.path.exists(abovePath):
    createDirs(abovePath)

  try:
    os.mkdir(path)
  except Exception: # OSError:
    # log.warning('failed to make {0} because {1}', path, e)
    pass


def copyfile(srcPath, dstDir):
  '''
    copies srcPath to dstPath, creating the directory structure if necessary for the destination
    srcPath: absolute file path
    dstDir:  absolute directory path
  '''

  if not os.path.exists(srcPath):
    return False

  dstPath = os.path.join(dstDir, os.path.basename(srcPath))

  if os.path.exists(dstPath):
    if not srcNewer(srcPath, dstPath):
      return

  # in case the path doesn't already exist
  createDirs(dstDir)

  shutil.copy2(srcPath, dstDir)

  log.debug('{0} copied to {1}', srcPath, dstPath)

  return True


def loadJsonFile(jsonPath):
  '''
    load a json config file
    NOTE: no check for existence of the path so that logging warnings can be controlled elsewhere
  '''
  if os.path.splitext(jsonPath)[1] != '.json':
    log.warning('{0} is not json', jsonPath)
    return None
  with open(jsonPath) as f:
    return json.loads(removeComments(f))


def removeComments(f):
  '''
    removes // style comments from a file, num of lines stays the same
  '''
  sansComments = ''
  inQuotes = False
  for l in f:
    i = 0
    for c in l:
      if c == '"':
        inQuotes = not inQuotes
      elif c == '/' and l[i + 1] == '/' and not inQuotes:
        sansComments += '\n'
        break
      i += 1
      sansComments += c
  return sansComments


def runCmd(cmd):
  '''
    runs a command and blocks until it's done, returns the output
  '''
  try:
    p = subprocess.Popen(cmd, stdout = subprocess.PIPE, stderr = subprocess.PIPE)
  except subprocess.CalledProcessError as e:
    return f('cmd failed: {0} because: {1}', ' '.join(cmd), e.output)
  except Exception:
    return f('cmd failed: {0}', ' '.join(cmd))
  stdout, stderr = p.communicate()
  output = ''
  if len(stderr):
    output += stderr.decode('utf-8')
  if len(stdout):
    output += stdout.decode('utf-8')
  return output




import os
from setuptools import setup
from setuptools.command.install import install as baseInstall

pybythecGlobals = '''
{
  "binaryFormat": "64bit",
  "buildType": "release",
  "installPath": ".",
  "buildDir": ".pybythec",
  "plusplus": true,
  "compiler": {
    "linux": "gcc",
    "macOs": "clang",
    "windows": "msvc"
  },
  "filetype": {
    "linux": "elf",
    "macOs": "mach-o",
    "windows": "pe"
  },
  "defines": {
    "msvc": [
      "NOMINMAX",
      "VC_EXTRALEAN",
      "WIN32_LEAN_AND_MEAN"
    ],
    "macOs": "__APPLE__"
  },
  "flags": {
    "gcc": {
      "all": [
        "-Wall",
        "-Wno-deprecated",
        "-fno-gnu-keywords"
      ],
      "debug": "-g",
      "release": "-O3",
      "linux": {
        "static": "-fPIC",
        "dynamic": "-fPIC",
        "plugin": "-fPIC"
      }
    },
    "clang": {
      "all": [
        "-Wall",
        "-Wno-deprecated",
        "-fno-gnu-keywords"
      ],
      "debug": "-g",
      "release": "-O3",
      "linux": {
        "static": "-fPIC",
        "dynamic": "-fPIC",
        "plugin": "-fPIC"
      }
    },
    "macOs": {
      "dynamic": "-fno-common",
      "plugin": "-fno-common"
    },
    "msvc": {
      "all": [
        "/EHsc",
        "/Gy",
        "/nologo",
        "/errorReport:prompt"
      ],
      "debug": {
        "all": [
          "/W1",
          "/RTC1",
          "/Z7"
        ]
      },
      "release": {
        "all": [
          "/DNDEBUG",
          "/O2",
          "/GL"
        ]
      },
      "dynamic": "/LD",
      "plugin": "/LD"
    }
  },
  "linkFlags": {
    "gcc": {
      "all": "-Wall",
      "release": "-O3",
      "exe": "-Wl,-rpath=$ORIGIN/",
      "dynamic": "-shared",
      "plugin": "-shared",
      "linux": {
        "dynamic": "-fPIC",
        "plugin": "-fPIC"
      }
    },
    "clang": {
      "all": "-Wall",
      "release": "-O3",
      "linux": {
        "exe": "-Wl,-rpath=$ORIGIN/",
        "dynamic": [
          "-shared",
          "-fPIC"
        ],
        "plugin": [
          "-shared",
          "-fPIC"
        ]
      },
      "macOs": {
        "dynamic": "-dynamiclib",
        "plugin": "-bundle"
      }
    },
    "msvc": {
      "all": [
        "/NOLOGO",
        "/ERRORREPORT:PROMPT"
      ],
      "debug": "/DEBUG",
      "release": [
        "/INCREMENTAL:NO",
        "/OPT:REF",
        "/OPT:ICF",
        "/LTCG"
      ]
    }
  },
  "libs": {
    "gcc": "pthread"
  },
  "msvc-default": "msvc-140",
  "bins": {
    "msvc-090": {
      "all": [
        "C:/Program Files (x86)/Microsoft Visual Studio 9.0/Common7/IDE",
        "C:/Program Files/Microsoft SDKs/Windows/v6.0A/bin"
      ],
      "32bit": "C:/Program Files (x86)/Microsoft Visual Studio 9.0/VC/bin",
      "64bit": [
        "C:/Program Files (x86)/Microsoft Visual Studio 9.0/VC/bin/amd64",
        "C:/Program Files (x86)/Microsoft Visual Studio 9.0/VC/bin/x86_amd64"
      ]
    },
    "msvc-100": {
      "all": [
        "C:/Program Files (x86)/Microsoft Visual Studio 10.0/Common7/IDE",
        "C:/Program Files/Microsoft SDKs/Windows/v7.1/bin"
      ],
      "32bit": "C:/Program Files (x86)/Microsoft Visual Studio 10.0/VC/bin",
      "64bit": [
        "C:/Program Files (x86)/Microsoft Visual Studio 10.0/VC/bin/amd64",
        "C:/Program Files (x86)/Microsoft Visual Studio 10.0/VC/bin/x86_amd64"
      ]
    },
    "msvc-110": {
      "all": [
        "C:/Program Files (x86)/Microsoft Visual Studio 11.0/Common7/IDE",
        "C:/Program Files (x86)/Microsoft SDKs/Windows/v7.1A/bin"
      ],
      "32bit": "C:/Program Files (x86)/Microsoft Visual Studio 11.0/VC/bin",
      "64bit": [
        "C:/Program Files (x86)/Microsoft Visual Studio 11.0/VC/bin/amd64",
        "C:/Program Files (x86)/Microsoft Visual Studio 11.0/VC/bin/x86_amd64"
      ]
    },
    "msvc-140": {
      "all": [
        "C:/Program Files (x86)/Microsoft Visual Studio 14.0/Common7/IDE",
        "C:/Program Files (x86)/Microsoft SDKs/Windows/v10.0A/bin/NETFX 4.6.1 Tools"
      ],
      "32bit": "C:/Program Files (x86)/Microsoft Visual Studio 14.0/VC/bin",
      "64bit": "C:/Program Files (x86)/Microsoft Visual Studio 14.0/VC/bin/amd64"
    }
  },
  "incPaths": {
    "msvc-090": [
      "C:/Program Files (x86)/Microsoft Visual Studio 9.0/VC/include",
      "C:/Program Files/Microsoft SDKs/Windows/v6.0A/Include"
    ],
    "msvc-100": [
      "C:/Program Files (x86)/Microsoft Visual Studio 10.0/VC/include",
      "C:/Program Files/Microsoft SDKs/Windows/v7.1/Include"
    ],
    "msvc-110": [
      "C:/Program Files (x86)/Microsoft Visual Studio 11.0/VC/include",
      "C:/Program Files (x86)/Microsoft SDKs/Windows/v7.1A/Include"
    ],
    "msvc-140": [
      "C:/Program Files (x86)/Microsoft Visual Studio 14.0/VC/include",
      "C:/Program Files (x86)/Windows Kits/10/Include/10.0.14393.0/ucrt",
      "C:/Program Files (x86)/Windows Kits/10/Include/10.0.14393.0/um",
      "C:/Program Files (x86)/Windows Kits/10/Include/10.0.14393.0/shared"
    ]
  },
  "libPaths": {
    "msvc-090": {
      "32bit": [
        "C:/Program Files (x86)/Microsoft Visual Studio 9.0/VC/lib",
        "C:/Program Files/Microsoft SDKs/Windows/v6.0A/Lib"
      ],
      "64bit": [
        "C:/Program Files (x86)/Microsoft Visual Studio 9.0/VC/lib/amd64",
        "C:/Program Files/Microsoft SDKs/Windows/v6.0A/Lib/x64"
      ]
    },
    "msvc-100": {
      "32bit": [
        "C:/Program Files (x86)/Microsoft Visual Studio 10.0/VC/lib",
        "C:/Program Files/Microsoft SDKs/Windows/v7.1/Lib"
      ],
      "64bit": [
        "C:/Program Files (x86)/Microsoft Visual Studio 10.0/VC/lib/amd64",
        "C:/Program Files/Microsoft SDKs/Windows/v7.1/Lib/x64"
      ]
    },
    "msvc-110": {
      "32bit": [
        "C:/Program Files (x86)/Microsoft Visual Studio 11.0/VC/lib",
        "C:/Program Files (x86)/Microsoft SDKs/Windows/v7.1A/Lib"
      ],
      "64bit": [
        "C:/Program Files (x86)/Microsoft Visual Studio 11.0/VC/lib/amd64",
        "C:/Program Files (x86)/Microsoft SDKs/Windows/v7.1A/Lib/x64"
      ]
    },
    "msvc-140": {
      "32bit": [
        "C:/Program Files (x86)/Microsoft Visual Studio 14.0/VC/lib",
        "C:/Program Files (x86)/Windows Kits/10/Lib/10.0.14393.0/ucrt/x86",
        "C:/Program Files (x86)/Windows Kits/10/Lib/10.0.14393.0/um/x86"
      ],
      "64bit": [
        "C:/Program Files (x86)/Microsoft Visual Studio 14.0/VC/lib/amd64",
        "C:/Program Files (x86)/Windows Kits/10/Lib/10.0.14393.0/ucrt/x64",
        "C:/Program Files (x86)/Windows Kits/10/Lib/10.0.14393.0/um/x64"
      ]
    }
  }
}
'''

class installer(baseInstall):

  def run(self):
    globalsPath = os.path.expanduser('~') + '/.pybythecGlobals.json'
    print('installing ' + globalsPath)
    with open(globalsPath, 'w') as f:
      f.write(pybythecGlobals)
    baseInstall.run(self)


description = 'A lightweight cross-platform build system for c/c++, written in python'

setup(
    name = 'pybythec',
    version = '0.9.43',
    author = 'glowtree',
    author_email = 'tom@glowtree.com',
    url = 'https://github.com/glowtree/pybythec',
    description = description,
    long_description = str(open('README.rst', 'r').read()).replace(description, ''),
    packages = ['pybythec'],
    entry_points = {
        'console_scripts': ['pybythec=pybythec.command_line:main'],
    },
    license = 'LICENSE',
    test_suite = 'test',
    cmdclass = {'install': installer}
)



import os
import platform
import subprocess
from pybythec import utils
from pybythec.utils import f
from pybythec.utils import PybythecError

log = utils.Logger('pybythec')


class BuildElements:

  def __init__(self,
               osType = None,
               compiler = None,
               buildType = None,
               binaryFormat = None,
               projConfigPath = None,
               globalConfigPath = None,
               projConfig = None,
               globalConfig = None,
               libDir = None):
    '''
      osType: operating system: currently linux, macOs, or windows
      builds: list of build variations
      compiler: any variation of gcc, clang, or msvc ie gcc-4.4, msvc110, if plusplus is true gcc, clang will become g++, clang++
      buildType: debug release etc
      binaryFormat: 32bit, 64bit etc
      projConfigPath: path to a pybythec project config file (json)
      globalConfigPath: path to a pybythec global config file (json)
      projConfig: dict of the project config
      globalConfig: dict of the global config
      libDir: directory path of the library being built, likely only used when building a library as a dependency (ie from a project)

      parses config files: global, project, local if they exist to determine build elements (state), uses function arguments as overrides
    '''

    self.localConfig = None
    self.projConfig = projConfig
    self.globalConfig = globalConfig
    self.libDir = libDir

    # overrides config files
    self.osTypeOverride = osType
    self.compilerOverride = compiler
    self.buildTypeOverride = buildType
    self.binaryFormatOverride = binaryFormat
    
    self.buildType = None
    self.version = None
    self.targetName = None # name of the target
    self.targetFilename = None # name of the target + extension

    self.builds = None  # a list of custom build keys
    self.osType = None  # linux, macOs, windows
    self.binaryType = None  # exe, static, dynamic, plugin
    
    self.version = 0
    self.binaryFormat = None # 32bit, 64bit etc
    self.libInstallPathAppend = True
    self.plusplus = True
    self.locked = False
    self.buildDir = '.'
    self.showCompilerCmds = False
    self.showLinkerCmds = False

    self.msvcDefault = None

    self.cwDir = os.getcwd()
    if self.libDir:
      self.cwDir = self.libDir
    
    # global config
    if not self.globalConfig:
      if globalConfigPath:
        if os.path.exists(globalConfigPath):
          self.globalConfig = utils.loadJsonFile(globalConfigPath)
        else:
          log.warning('{0} doesn\'t exist', globalConfigPath)
      elif 'PYBYTHEC_GLOBALS' in os.environ:
        globalConfigPath = os.environ['PYBYTHEC_GLOBALS']
        if os.path.exists(globalConfigPath):
          self.globalConfig = utils.loadJsonFile(globalConfigPath)
        else:
          log.warning('PYBYTHEC_GLOBALS points to {0}, which doesn\'t exist', globalConfigPath)
      elif os.path.exists('.pybythecGlobals.json'):
        self.globalConfig = utils.loadJsonFile('.pybythecGlobals.json')
      elif os.path.exists('pybythecGlobals.json'):
        self.globalConfig = utils.loadJsonFile('pybythecGlobals.json')
      else: # check the home directory
        homeDirPath = ''
        if platform.system() == 'Windows':
          homeDirPath = os.environ['USERPROFILE']
        else:
          homeDirPath = os.environ['HOME']
        if os.path.exists(homeDirPath + '/.pybythecGlobals.json'):
          self.globalConfig = utils.loadJsonFile(homeDirPath + '/.pybythecGlobals.json')
        elif os.path.exists(homeDirPath + '/pybythecGlobals.json'):
          self.globalConfig = utils.loadJsonFile(homeDirPath + '/pybythecGlobals.json')
        else: # end of the line
          log.warning('no pybythecGlobals.json found in the home directory (hidden or otherwise)')
    if not self.globalConfig:
      log.warning('not using a global configuration')
    
    # project config
    if not self.projConfig:
      if projConfigPath:
        if os.path.exists(projConfigPath):
          self.projConfig = utils.loadJsonFile(projConfigPath)
        else:
          log.warning('{0} doesn\'t exist', projConfigPath)
      elif 'PYBYTHEC_PROJECT' in os.environ:
        projConfPath = os.environ['PYBYTHEC_PROJECT']
        if os.path.exists(projConfPath):
          self.projConfig = utils.loadJsonFile(projConfPath)
        else:
          log.warning('PYBYTHEC_PROJECT points to {0}, which doesn\'t exist', projConfPath)
      else:
        if os.path.exists(self.cwDir + '/pybythecProject.json'):
          self.projConfig = utils.loadJsonFile(self.cwDir + '/pybythecProject.json')
        elif os.path.exists(self.cwDir + '/.pybythecProject.json'):
          self.projConfig = utils.loadJsonFile(self.cwDir + '/.pybythecProject.json')
    # if not projConfig:
      # log.warning('not using a project pybythec configuration')

    # local config, expected to be in the current working directory
    self.localConfig = None
    localConfigPath = self.cwDir + '/pybythec.json'
    if not os.path.exists(localConfigPath):
      localConfigPath = self.cwDir + '/.pybythec.json'
    if os.path.exists(localConfigPath):
      self.localConfig = utils.loadJsonFile(localConfigPath)

    #
    # first iteration to get osType and custom keys (right now just for the compiler)
    #
    if self.globalConfig is not None:
      self._getBuildElements1(self.globalConfig)
    if self.projConfig is not None:
      self._getBuildElements1(self.projConfig)
    if self.localConfig is not None:
      self._getBuildElements1(self.localConfig)

    # self.targetName = self.target
    self.targetFilename = self.targetName

    if self.osTypeOverride:
      self.osType = self.osTypeOverride

    if self.osType:
      if self.osType not in ['linux', 'macOs', 'windows']: # validate
        log.warning('{0} invalid osType, defaulting to the native os', self.osType)
        self.osType = None
    if not self.osType: # use the native os
      if platform.system() == 'Linux':
        self.osType = 'linux'
      elif platform.system() == 'Darwin':
        self.osType = 'macOs'
      elif platform.system() == 'Windows':
        self.osType = 'windows'
      else:
        raise PybythecError('os needs to be linux, macOs or windows')


  def configBuild(self, buildName = None):
    '''
    '''
    # set by the config files first
    self.compiler = None  # g++-4.4 g++ clang++ msvc-110 etc
    self.filetype = None  # elf, mach-o, pe
    self.installPath = None
  
    self.sources = []
    self.libs = []
    self.defines = []
    self.flags = []
    self.linkFlags = []

    self.incPaths = []
    self.extIncPaths = []  # these will not be checked for timestamps
    self.libPaths = []
    self.libSrcPaths = []

    self.qtClasses = []


    # 2 keys at this point for a potentially nested compiler: osType and buildName
    keys = [self.osType]
    if buildName:
      keys.append(buildName)

    #
    # second iteration to get the other configs that can't be nested (the compiler being the exception)
    #
    if self.globalConfig is not None:
      self._getBuildElements2(self.globalConfig, keys)
    if self.projConfig is not None:
      self._getBuildElements2(self.projConfig, keys)
    if self.localConfig is not None:
      self._getBuildElements2(self.localConfig, keys)

    # compiler stuff
    if self.compilerOverride:
      self.compiler = self.compilerOverride

    if not self.compiler:
      raise PybythecError('compiler not found')

    # validate compiler and  determine root: can be gcc, clang or msvc
    self.compilerRoot = None
    if self.compiler.startswith('gcc') or self.compiler.startswith('g++'):
      self.compilerRoot = 'gcc'
    elif self.compiler.startswith('clang') or self.compiler.startswith('clang++'):
      self.compilerRoot = 'clang'
    elif self.compiler.startswith('msvc'):
      self.compilerRoot = 'msvc'
      if self.compiler == 'msvc': # needs a version ie msvc-110 for pathing
        if self.msvcDefault:
          self.compiler = self.msvcDefault
        else:
          raise PybythecError('msvc has no default set, try setting the compiler to a specific version ie msvc-140')
        log.i('using {0}', self.compiler)
    else:
      raise PybythecError('unrecognized compiler {0}, using the default based on osType', self.compiler)


    if self.buildTypeOverride:
      self.buildType = self.buildTypeOverride

    if self.binaryFormatOverride:
      self.binaryFormat = self.binaryFormatOverride

    keys += ['all', self.compilerRoot, self.compiler, self.binaryType, self.buildType, self.binaryFormat]

    if self.globalConfig is not None:
      self._getBuildElements3(self.globalConfig, keys)
    if self.projConfig is not None:
      self._getBuildElements3(self.projConfig, keys)
    if self.localConfig is not None:
      self._getBuildElements3(self.localConfig, keys)

    # deal breakers (that don't appear in the default pybythecGlobals.json)
    if not self.targetName:
      raise PybythecError('no target specified')
    if not self.binaryType:
      raise PybythecError('no binary type specified')
    if self.binaryType not in ('exe', 'static', 'dynamic', 'plugin'):
      raise PybythecError('unrecognized binary type: ' + self.binaryType)
    if not self.sources:
      raise PybythecError('no source files specified')


    #
    # compiler config
    #
    self.compilerCmd = None
    self.linker = None
    self.targetFlag = None
    self.libFlag = None
    self.libPathFlag = None
    self.objExt = None
    self.objPathFlag = None
    self.staticExt = None
    self.dynamicExt = None
    self.pluginExt = None

    #
    # gcc / clang
    #
    if self.compilerRoot == 'gcc' or self.compilerRoot == 'clang':
      if self.plusplus:
        if self.compiler.startswith('gcc'):
          self.compiler = self.compiler.replace('gcc', 'g++')
        elif self.compiler.startswith('clang-') or self.compiler == ('clang'):
          self.compiler = self.compiler.replace('clang', 'clang++')

      self.compilerCmd = self.compiler
      self.objFlag = '-c'
      self.objExt = '.o'
      self.objPathFlag = '-o'
      self.defines.append('_' + self.binaryFormat.upper())  # TODO: you sure this is universal?

      # link
      self.linker = self.compilerCmd  # 'ld'
      self.targetFlag = '-o'
      self.libFlag = '-l'
      self.libPathFlag = '-L'
      self.staticExt = '.a'
      self.dynamicExt = '.so'
      self.pluginExt = '.so'

      if self.filetype == 'mach-o':
        self.dynamicExt = '.dylib'
        self.pluginExt = '.bundle'

      if self.binaryType == 'static' or self.binaryType == 'dynamic':
        self.targetFilename = 'lib' + self.targetName

      if self.binaryType == 'exe':
        pass
      elif self.binaryType == 'static':
        self.targetFilename = self.targetFilename + '.a'
        self.linker = 'ar'
        self.targetFlag = 'r'
      elif self.binaryType == 'dynamic':
        self.targetFilename = self.targetFilename + self.dynamicExt
      elif self.binaryType == 'plugin':
        self.targetFilename = self.targetFilename + self.pluginExt

    #
    # msvc / msvc
    #
    elif self.compilerRoot == 'msvc':

      # compile
      self.compilerCmd = 'cl'
      self.objFlag = '/c'
      self.objExt = '.obj'
      self.objPathFlag = '/Fo'

      # link
      self.linker = 'link'
      self.targetFlag = '/OUT:'
      self.libFlag = ''
      self.libPathFlag = '/LIBPATH:'
      self.staticExt = '.lib'
      self.dynamicExt = '.dll'
      if self.binaryFormat == '64bit':
        self.linkFlags.append('/MACHINE:X64')

      if self.binaryType == 'exe':
        self.targetFilename = self.targetName + '.exe'
      elif self.binaryType == 'static':
        self.targetFilename = self.targetName + self.staticExt
        self.linker = 'lib'
      elif self.binaryType == 'dynamic' or self.binaryType == 'plugin':
        self.targetFilename = self.targetName + self.dynamicExt
        self.linkFlags.append('/DLL')

    else:
      raise PybythecError('unrecognized compiler root: {0}', self.compilerRoot)


    # make sure the compiler is in PATH
    try:
      subprocess.call(self.compilerCmd, stdout = subprocess.PIPE, stderr = subprocess.PIPE)
    except OSError:
      raise PybythecError('compiler {0} is not found in PATH', self.compilerCmd)

    # make sure the linker is in PATH
    try:
      subprocess.call(self.linker, stdout = subprocess.PIPE, stderr = subprocess.PIPE)
    except OSError:
      raise PybythecError('linker {0} is not found in PATH', self.linker)

    #
    # determine paths
    #
    self.installPath = utils.makePathAbsolute(self.cwDir, self.installPath)
    self._resolvePaths(self.cwDir, self.sources)
    self._resolvePaths(self.cwDir, self.incPaths)
    self._resolvePaths(self.cwDir, self.extIncPaths)
    self._resolvePaths(self.cwDir, self.libPaths)
    self._resolvePaths(self.cwDir, self.libSrcPaths)

    self.binaryRelPath = f('/{0}/{1}/{2}', self.buildType, self.compiler, self.binaryFormat)

    binRelPath = self.binaryRelPath
    if buildName:
      binRelPath += '/' + buildName

    self.buildPath = utils.makePathAbsolute(self.cwDir, './' + self.buildDir + binRelPath)

    # if self.libInstallPathAppend and (self.binaryType == 'static' or self.binaryType == 'dynamic'):
    if self.libInstallPathAppend and (self.binaryType in ['static', 'dynamic']):
      self.installPath += binRelPath

    self.targetInstallPath = os.path.join(self.installPath, self.targetFilename)

    self.infoStr = f('{0} ({1} {2} {3}', self.targetName, self.buildType, self.compiler, self.binaryFormat)
    if buildName:
      self.infoStr += ' ' + buildName
    self.infoStr += ')'



  def _getBuildElements1(self, configObj):
    '''
      elements that aren't nested
    '''
    if 'buildType' in configObj:
      self.buildType = os.path.expandvars(configObj['buildType'])

    if 'version' in configObj:
      self.version = os.path.expandvars(configObj['version'])

    if 'target' in configObj:
      self.targetName = os.path.expandvars(configObj['target'])

    if 'builds' in configObj:
      self.builds = configObj['builds']

    if 'osType' in configObj:
      self.osType = os.path.expandvars(configObj['osType'])

    if 'binaryType' in configObj:
      self.binaryType = os.path.expandvars(configObj['binaryType'])

    if 'binaryFormat' in configObj:
      self.binaryFormat = os.path.expandvars(configObj['binaryFormat'])

    if 'libInstallPathAppend' in configObj:
      self.libInstallPathAppend = configObj['libInstallPathAppend']

    if 'plusplus' in configObj:
      self.plusplus = configObj['plusplus']

    if 'locked' in configObj:
      self.locked = configObj['locked']

    if 'buildDir' in configObj:
      self.buildDir = configObj['buildDir']

    if 'showCompilerCmds' in configObj:
      self.showCompilerCmds = configObj['showCompilerCmds']

    if 'showLinkerCmds' in configObj:
      self.showLinkerCmds = configObj['showLinkerCmds']

    if 'msvc-default' in configObj:
      self.msvcDefault = configObj['msvc-default']


  def _getBuildElements2(self, configObj, keys = []):
    '''
      elements that are nested in a finite / special case way, currently just the compiler
    '''
    # compiler can be nested in a dict with 2 valid key types: osType and a build name
    if 'compiler' in configObj:
      compilerList = []
      self._getArgsList(compilerList, configObj['compiler'], keys)
      if len(compilerList):
        self.compiler = compilerList[0]
        if len(compilerList) > 1:
          log.warning('couldn\'t resolve to single compiler, compiler options: {0}, selecting {1}', compilerList, self.compiler)


  def _getBuildElements3(self, configObj, keys = []):
    '''
      elements that are potentially nested in any which way
    '''
    separartor = ':'
    if platform.system() == 'Windows':
      separartor = ';'

    # TODO: PATH will grow for any build with dependencies, is there a way to prevent it?
    if 'bins' in configObj:
      bins = []
      self._getArgsList(bins, configObj['bins'], keys)
      for bin in bins:
        os.environ['PATH'] = bin + separartor + os.environ['PATH']

    if 'sources' in configObj:
      self._getArgsList(self.sources, configObj['sources'], keys)

    if 'libs' in configObj:
      self._getArgsList(self.libs, configObj['libs'], keys)

    if 'defines' in configObj:
      self._getArgsList(self.defines, configObj['defines'], keys)

    if 'flags' in configObj:
      self._getArgsList(self.flags, configObj['flags'], keys)

    if 'linkFlags' in configObj:
      self._getArgsList(self.linkFlags, configObj['linkFlags'], keys)

    if 'incPaths' in configObj:
      self._getArgsList(self.incPaths, configObj['incPaths'], keys)

    if 'extIncPaths' in configObj:
      self._getArgsList(self.extIncPaths, configObj['extIncPaths'], keys)

    if 'libPaths' in configObj:
      self._getArgsList(self.libPaths, configObj['libPaths'], keys)

    if 'libSrcPaths' in configObj:
      self._getArgsList(self.libSrcPaths, configObj['libSrcPaths'], keys)

    if 'qtClasses' in configObj:
      self._getArgsList(self.qtClasses, configObj['qtClasses'], keys)

    if 'filetype' in configObj:
      filetypes = []
      self._getArgsList(filetypes, configObj['filetype'], keys)
      if len(filetypes):
        self.filetype = filetypes[0]

    if 'installPath' in configObj:
      installPaths = []
      self._getArgsList(installPaths, configObj['installPath'], keys)
      if len(installPaths):
        self.installPath = installPaths[0]

  def _resolvePaths(self, absPath, paths):
    '''
    '''
    i = 0
    for path in paths:
      paths[i] = utils.makePathAbsolute(absPath, path)
      i += 1

  def _getArgsList(self, argsList, args, keys = []):
    '''
      recursivley parses args and appends it to argsList if it has any of the keys
      args can be a dict, str (space-deliminated) or list
    '''
    if type(args) == dict:
      for key in keys:
        if key in args:
          self._getArgsList(argsList, args[key], keys)
    else:
      if type(args) == str or type(args).__name__ == 'unicode':
        argsList.append(os.path.expandvars(args))
      elif type(args) == list:
        for arg in args:
          argsList.append(os.path.expandvars(arg))
